#include "LocBra.h"
#include "LocBra1.h"

static int
  CPXPUBLIC mycallback (CPXENVptr env, void *cbdata, int wherefrom, void *cbhandle);

int main(int argc, char **argv)
{
   int        i, nIter, maxIter, status, Improved, lpstat, div;
   int        first, nSolLim, type, refine, nNodes, sol;
   double     Time, timeLimit, time4Neig, t4N, t_end, addTime; 
   char       fname[40];
   FILE       *para;
   double     xBarValue, bestValue, rhs;
 
   double    *xBar;
   char      *ctype;

   if (argc != 7) {
      printf("\n Usage: %s totalTime time4N kSize sol istName", argv[0]);
      printf("\n      where:     totalTime  -> given TimeLimit");
      printf("\n                    time4N  -> time for Neighborhood");
      printf("\n                   addTime  -> additional time at the root node");
      printf("\n                         k  -> size");
      printf("\n                       sol  -> starting sol: NO = 0, YES, from file = 1");
      printf("\n                       sol  -> starting sol: YES, stupid = 2, YES, rounding = 3");
      printf("\n                   istName  -> instance name");
      printf("\n");
      exit (0);
   }
   sscanf(argv[1], "%lf", &timeLimit);
   sscanf(argv[2], "%lf", &time4Neig);
   sscanf(argv[3], "%lf", &addTime);
   sscanf(argv[4], "%d",  &kSize);
   sscanf(argv[5], "%d",  &sol);
   sscanf(argv[6], "%s",  fname);

   { 
    int    dm;
    para = fopen("param.txt", "r");
    fscanf(para, "%d\n", &dm); emphasis = dm;
    fscanf(para, "%d\n", &dm); presolve = dm;
    fscanf(para, "%d\n", &dm); heurFreq = dm;
    fscanf(para, "%d\n", &dm); prec = dm;
    fscanf(para, "%d\n", &dm); mipInt = dm;
    fscanf(para, "%d\n", &dm); nIntervals = dm;
    fscanf(para, "%d\n", &dm); original = dm;
    fscanf(para, "%d\n", &dm); exact = dm;
    fscanf(para, "%d\n", &dm); video = dm;
    fclose(para);
   }

   printf("\n LB timeLimit, time4Neig, addTime, emphasis, presolve, heurFreq");
   printf(" prec, mipInt, timeInt, kSize, original, exact, video, sol, fname");
   printf("\n LB %10.0f %10.0f %10.0f %d %d %d %d %d %d %d %d %d %d %d %s \n \n", 
              timeLimit, time4Neig, addTime, emphasis, presolve, heurFreq, 
	      prec, mipInt, nIntervals, kSize, original, exact, video, sol, fname);
   fflush(stdout);

/* initialize environment variables */
   env   = NULL;
   lp    = NULL;
   Eps   = .0000001;
   Inf   = 1000000000; 
   dInf  = 100000000000.; 
   slack = 1000000.;
   nI = 9999; flag = 0;

/* CPLEX initialization */
   env = CPXopenCPLEX(&status);
   if (env == NULL) printf("\n error CPXopenCPLEX");
   lp = CPXcreateprob(env, &status, "LocBra");
   if (lp == NULL) printf("\n error CPXcreateprob");

/* read the problem */
   status = CPXreadcopyprob(env, lp, fname, NULL);
   if (status) printf("\n error CPXreadcopyprob, status = %d", status);

/* get information about the problem */
   MinMax  = CPXgetobjsen(env, lp);
   ncols = CPXgetnumcols(env, lp);
   nrows = CPXgetnumrows(env, lp);

#ifdef VERBOSE
   printf("\n Problem info: MinMax = %d, ncols = %d, nrows = %d\n",MinMax, ncols, nrows);
   fflush(stdout);
#endif

/* cplex paramenters */
   status = setup_cplex();
   if (status) { printf ("\nError in setup_cplex: %d", status); exit(1); }

/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
   INITIAL SOLUTION and (possibly) MODEL RELAXATION */
   if (sol) {
      int     ninf, j, cnt;
      int     feas = 0;
      double  val;
      FILE    *input;
      int     *inf, *indices;
      double  *initial, *lb, *db, *ub;
      char    *lu, *ct;
      char    name[20];
      initial = (double*)calloc(ncols, sizeof(double));

      switch (sol) {
      case 1: /* an initial solution either feasible or infeasible is given in the file `solution.mst' */

              cnt = 0;
	      lb = (double*)calloc(ncols, sizeof(double));
	      db = (double*)calloc(ncols, sizeof(double));
	      ub = (double*)calloc(ncols, sizeof(double));
              lu = (char*)calloc(ncols, sizeof(char));
              indices = (int*)calloc(ncols, sizeof(int));
              ct = (char*)calloc(ncols, sizeof(char));
              status = CPXgetctype(env, lp, ct, 0, ncols - 1);
              input = fopen("solution.mst", "r");
	      fscanf(input,"%s\n", name);
              fscanf(input,"%s", name);
	      do {
                  fscanf(input,"%lf\n", &val);
                  status = CPXgetcolindex(env, lp, name, &j);
		  initial[j] = val;
		  db[cnt] = val; indices[cnt] = j; cnt++;
                  fscanf(input,"%s", name);
              } while(name[0] != 'E' || name[1] != 'N' || name[2] != 'D' ||
		      name[3] != 'A' || name[4] != 'T' || name[5] != 'A'); //"ENDATA");
              fclose(input);
	      for (j = 0; j < cnt; j++) lu[j] = 'B';
	      status = CPXgetlb(env, lp, lb, 0, ncols-1);
	      status = CPXgetub(env, lp, ub, 0, ncols-1);
	      status = CPXchgbds(env, lp, cnt, indices, lu, db);
              status = CPXmipopt(env, lp); status = CPXflushstdchannels(env);
	      lpstat = CPXgetstat(env, lp);
	      printf("\n status of the initial solution provided = %d", lpstat); fflush(stdout);
	      if (lpstat == 101 || lpstat == 102) {
		 printf("\n initial solution provided is FEASIBLE");
		 status = CPXgetmipx(env, lp, initial, 0, ncols - 1);
		 feas = 1;
	      } else {
		 printf("\n initial solution provided is INFEASIBLE");
		 for (j = 0; j < ncols; j++) if (ct[j] == 'C') initial[j] = lb[j];
              }
	      for (j = 0; j < ncols; j++) { indices[j] = j; lu[j] = 'L'; }
	      status = CPXchgbds(env, lp, ncols, indices, lu, lb);
	      for (j = 0; j < ncols; j++) lu[j] = 'U';
	      status = CPXchgbds(env, lp, ncols, indices, lu, ub);
	      free(lb); free(indices); free(lu); free(ct); free(db); free(ub);
              break;

      case 2: /* an initial stupid solution is computed */

              lb = (double*)calloc(ncols, sizeof(double));
              status = CPXgetlb(env, lp, lb, 0, ncols-1);
              for (j = 0; j < ncols; j++) initial[j] = lb[j];
              free(lb);
              break;

      case 3: /* an initial solution based on rounding is computed */
 
              ct = (char*)calloc(ncols, sizeof(char));
              status = CPXgetctype(env, lp, ct, 0, ncols - 1);
              lb = (double*)calloc(ncols, sizeof(double));
              status = CPXgetlb(env, lp, lb, 0, ncols-1);

              status = CPXchgprobtype(env, lp, CPXPROB_LP);
              if (status) printf("\n CPXchgprobtype: %d", status);
              status = CPXprimopt(env, lp); status = CPXflushstdchannels(env);
              if (status) printf("\n CPXprimopt: %d", status);
              status = CPXgetx(env, lp, initial, 0, ncols - 1);
              if (status) printf("\n CPXgetx: %d", status);

              for (j = 0; j < ncols; j++) {
                  if (ct[j] != 'C') {
                     double dummy;
                     dummy = floor(initial[j]);
                     if (initial[j] - dummy > .5) initial[j] = dummy + 1;
                     else                         initial[j] = dummy;
                  } else initial[j] = lb[j];
              }

              status = CPXchgprobtype(env, lp, CPXPROB_MILP);
              if (status) printf("\n CPXchgprobtype: %d", status);
              indices = (int*)calloc(ncols, sizeof(int));
	      for (j = 0; j < ncols; j++) indices[j] = j;
	      status = CPXchgctype (env, lp, ncols, indices, ct);

              free(lb); free(ct); free(indices);
              break;
      }  

      /* testing feasibility */
      if (feas) goto down;
      inf = (int*)calloc(nrows, sizeof(int));
      ninf = 0;
      for (i = 0; i < nrows; i++) {
          int    nnz, su;
          double cRhs, rRhs;
          char   se;
          int    *rmbeg, *rmind;
          double *rmval;
          cRhs = 0.;
          rmbeg = (int*)calloc(2, sizeof(int));
          rmind = (int*)calloc(ncols, sizeof(int));
          rmval = (double*)calloc(ncols, sizeof(double));
          status = CPXgetrows(env, lp, &nnz, rmbeg, rmind, rmval, ncols, &su, i, i);
          for (j = 0; j < nnz; j++) cRhs += rmval[j] * initial[rmind[j]];
          status = CPXgetsense(env, lp, &se, i, i);
          status = CPXgetrhs(env, lp, &rRhs, i, i);
          switch(se) {
          case 'L':
                    if (cRhs - Eps > rRhs) { inf[i] = 1; ninf++; }
                    break;
          case 'E':
                    if (fabs(cRhs-rRhs) > Eps) { inf[i] = 2; ninf++; }
                    break;
          case 'G':
                    if (cRhs + Eps < rRhs) { inf[i] = -1; ninf++; }
                    break;
          }
          free(rmbeg); free(rmind); free(rmval);
      }
      if (ninf) {
         /* model relaxation REQUIRED */
         printf("\n model relaxation REQUIRED"); fflush(stdout);
         int nc = 0;
         for (i = 0; i < nrows; i++) {
             switch(inf[i]) {
             case -1: 
                     {
                      int    cmind;
                      double obj, cmval;
                      int    *cmbeg;
                      obj = slack;
                      cmbeg = (int*)calloc(2, sizeof(int));
                      cmbeg[0] = 0; cmind = i; cmval = 1.; 
                      status = CPXaddcols(env, lp, 1, 1, &obj, cmbeg, &cmind, &cmval, NULL, NULL, NULL);
                      free(cmbeg);
                     }
                      break;
             case 0:
                      break;
             case 1:
                     {
                      int    cmind;
                      double obj, cmval;
                      int    *cmbeg;
                      obj = slack;
                      cmbeg = (int*)calloc(2, sizeof(int));
                      cmbeg[0] = 0; cmind = i; cmval = -1.; 
                      status = CPXaddcols(env, lp, 1, 1, &obj, cmbeg, &cmind, &cmval, NULL, NULL, NULL);
                      free(cmbeg);
                     }
                      break;
             case 2:
                     {
                      int    *cmind;
                      double *obj, *cmval;
                      int    *cmbeg;
                      obj   = (double*)calloc(2, sizeof(double));
                      cmbeg = (int*)calloc(2, sizeof(int));
                      cmind = (int*)calloc(2, sizeof(int));
                      cmval = (double*)calloc(2, sizeof(double));
                      cmbeg[0] = 0; cmbeg[1] = 1;
                      cmind[0] = cmind[1] = i;
                      cmval[0] = 1.; cmval[1] = -1.;
                      obj[0] = obj[1] = slack;
                      status = CPXaddcols(env, lp, 2, 2, obj, cmbeg, cmind, cmval, NULL, NULL, NULL);
                      free(cmbeg); free(obj); free(cmval); free(cmind);
                      nc++;
                     }
                      break;
             }
         }

         printf("\n %d infeasibilities (%d new cols), final number of COLS = %d\n", 
                    ninf, ninf + nc, ncols + ninf + nc); fflush(stdout);

         if (ninf + nc + ncols != CPXgetnumcols(env, lp)) { 
            printf("\n wrong column counting: %d %d %d %d\n", ninf, nc, ncols, CPXgetnumcols(env,lp));
            exit(1);
         } else {
           ncols += ninf + nc;
           nNewCols = ninf + nc;
         }

         if (sol > 1) wdummysol(initial);

      } else nNewCols = 0;

      free(inf);
down:
      free(initial);

#ifdef PRINT
      status = CPXwriteprob(env, lp, "relax.lp", "LP");
      exit(0);
#endif

      status = CPXsetintparam(env, CPX_PARAM_MIPSTART, 1);
      status = CPXreadcopymipstart(env, lp, "solution.mst"); 

   } else nNewCols = 0;
/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ */

/* optimization parameters */
   if (exact) dmax = 0;
   else       dmax = Inf;
   maxIter = Inf;
   div     = 0;
   first   = 1;
   refine  = 0;
   debug   = 0;
   if (timeLimit < 0.) timeLimit = dInf;
   timeSlot = timeLimit / nIntervals;
   slot = (int*)calloc(nIntervals, sizeof(int)); slot[0] = 1;

/* memory allocation */
   xBar = (double*)calloc(ncols, sizeof(double));
   ref = (int*)calloc(ncols, sizeof(int));
   ctype = (char*)calloc(ncols, sizeof(char));
   status = CPXgetctype(env, lp, ctype, 0, ncols - 1);

   nBcols = nIcols = nCcols = 0; 
   for (i = 0; i < ncols; i++) {
       if (ctype[i] == 'B') nBcols++;
       if (ctype[i] == 'I') nIcols++;
       if (ctype[i] == 'C') nCcols++;
   }

#ifdef VERBOSE
   printf("\n Problem info: MinMax = %d, ncols = %d, nrows = %d,",MinMax, ncols, nrows);
   printf(" nBcols = %d, nIcols = %d, nCcols = %d\n", nBcols, nIcols, nCcols);
   fflush(stdout);
#endif

/* initialization */
   nIter = dv = 0; Optimum = 0;
   t_start = second(); Time = 0.;
   overallBestValue = bestValue = (double)MinMax * dInf;
   t4N = dInf;

/* cplex paramenters */
   if (prec) {
      status = CPXsetdblparam(env, CPX_PARAM_EPAGAP, 1e-12);
      status = CPXsetdblparam(env, CPX_PARAM_EPGAP, 1e-06);
   }
   status = CPXsetintparam(env, CPX_PARAM_SCRIND, video);
   status = CPXsetintparam(env, CPX_PARAM_MIPINTERVAL, mipInt);
   status = CPXsetintparam(env, CPX_PARAM_HEURFREQ, heurFreq);
   if (emphasis >= 0) status = CPXsetintparam(env, CPX_PARAM_MIPEMPHASIS, emphasis);
   status = CPXsetintparam(env, CPX_PARAM_PREIND, presolve);

   status = CPXsetmipcallbackfunc(env, mycallback, NULL);
   if (status) {
      printf ("\nError in CPXsetlpcallbackfunc: %d", status);
      exit(1);
   }

   while (Time < timeLimit && nIter < maxIter && !Optimum) {
         printf("\n CPLEX call # %d started!\n", nIter + 1); fflush(stdout);

/* set TIME limit for each neighborhood exploration */
         if (t4N > timeLimit - Time) t4N = timeLimit - Time;
         status = CPXsetdblparam(env, CPX_PARAM_TILIM, t4N); 

/* set first/best neighborhood improvement option */
         nSolLim = Inf;
         if (first) nSolLim = 1;
         status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, nSolLim);

/* mipopt */
         status = CPXmipopt(env, lp);
         if (first) {
            status = CPXsetdblparam(env, CPX_PARAM_TILIM, addTime);
            status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, Inf);
            status = CPXmipopt(env, lp);
            status = CPXsetintparam(env, CPX_PARAM_MIPSTART, 0);
         }
	 status = CPXflushstdchannels(env);

         if (status) printf("\nError in CPXmipopt: %d", status);
         t4N = time4Neig;

/* check lpstat */
         if ((lpstat = CPXgetstat(env, lp))) {
            printf("\n   Optimization status = %d", lpstat);
            fflush(stdout);
         } else { printf("\nSomething wrong: lpstat = %d", lpstat); exit(0); }

/* 1 */  if (lpstat == 101 || lpstat == 102) {
            xBarValue = getSolution(xBar);
            Improved = checkSol(xBarValue, &bestValue, xBar); 
	    if (Optimum) goto end;
            if (!Improved) {
               printf("\n   an improvement was expected: something wrong!");
               exit(1);
            }
            printf("\n   IMPROVED solution = %16.4f", bestValue); fflush(stdout);
            if (!nIter) { Optimum = 1; goto end; }
            switch(MinMax) {
            case 1:  /* MINimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
		     break;
	    case -1: /* MAXimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTLO, bestValue + Eps);
		     break;
	    }
            div = first = 0; rhs = kSize; 
            type = 0; modifyRow(type);
          }    
/* 2 */  if (lpstat == 103) {
            if (!nIter) { 
               Optimum = 1; 
               printf("\n   NO FEASIBLE solution exists"); fflush(stdout);
               goto end; 
            }
            type = 0; modifyRow(type);   
            if (div || exact) { bestValue = (double)MinMax * dInf; t4N = dInf; dv++; first = 1; }
            switch(MinMax) {
            case 1:  /* MINimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
		     break;
	    case -1: /* MAXimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTLO, bestValue + Eps);
		     break;
	    }
            div = 1; rhs += kSize/2;
         }
/* 3 */  if (lpstat == 104 || lpstat == 107) {
            xBarValue = getSolution(xBar);
            if (nIter) {
               if (lpstat == 104) {
                  type = 2; modifyRow(type); 
               } else {
                  if (lpstat == 107) {
                     type = 1; modifyRow(type); 
                  } else { 
                     printf("\n strange lpstat"); exit(0); 
                  }
               }
            } else {
               status = CPXgetbestobjval(env, lp, &bestBound);
               if (status) printf("\nError in CPXgetbestobjval: %d", status);
               nNodes = CPXgetnodecnt(env, lp);
               printf("\n   Final bound = %16.4f after %d B&B nodes", bestBound, nNodes); fflush(stdout);
            }
            if (ncols != nBcols) {
               printf("\n   IMPROVED solution before REFINEMENT = %16.4f\n", xBarValue); fflush(stdout);
               refineSolution(&xBarValue, xBar);
            }
            Improved = checkSol(xBarValue, &bestValue, xBar);
	    if (Optimum) goto end;
            if (!Improved) {
               printf("\n   an improvement was expected: something wrong!");
               exit(1);
            }
            printf("\n   IMPROVED solution = %16.4f", bestValue); fflush(stdout);
            switch(MinMax) {
            case 1:  /* MINimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
		     break;
	    case -1: /* MAXimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTLO, bestValue + Eps);
		     break;
	    }
            div = first = 0; rhs = kSize;
         }
/* 4 */  if (lpstat == 108) {
            if (nIter && !div) { type = 2; modifyRow(type); } 
            else               { type = 1; modifyRow(type);
                                 if (!nIter) { printf("\n time limit and no sol!"); exit(0); } }
               
            if (div || exact) { bestValue = (double)MinMax * dInf; t4N = dInf; dv++; first = 1; rhs += kSize/2; } 
            else rhs -= kSize/2;
            switch(MinMax) {
            case 1:  /* MINimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
		     break;
	    case -1: /* MAXimization version */
		     status = CPXsetdblparam(env, CPX_PARAM_CUTLO, bestValue + Eps);
		     break;
	    }
            div = 1;
         }
        
         if (dv > dmax) { exact = 1; goto out; }

         if (debug == 1) status = CPXwriteprob(env, lp, "prima.lp", "LP");
         localBranch(ctype, xBar, refine, rhs);
         if (debug == 1) status = CPXwriteprob(env, lp, "dopo.lp", "LP");

end:
         t_end = second(); Time = t_end - t_start;
         printf("\n c. %16.4f - best %16.4f (%4d) - elapsed time %8.1f (%d)\n", 
                 bestValue, overallBestValue, nI, Time, dv); fflush(stdout);
         nIter++;
   }

out:
   if (exact && !Optimum) {
      t_end = second(); Time = t_end - t_start;
      printf("\n   Final locBra sol %16.4f - elapsed time %8.1f \n", overallBestValue, Time); fflush(stdout);
      status = CPXsetdblparam(env, CPX_PARAM_CUTUP, overallBestValue - Eps);
      switch(MinMax) {
      case 1:  /* MINimization version */
               status = CPXsetdblparam(env, CPX_PARAM_CUTUP, overallBestValue - Eps);
               break;
      case -1: /* MAXimization version */
               status = CPXsetdblparam(env, CPX_PARAM_CUTLO, overallBestValue + Eps);
               break;
      }
      status = CPXsetdblparam(env, CPX_PARAM_TILIM, timeLimit - Time);
      status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, Inf);
      status = CPXmipopt(env, lp); status = CPXflushstdchannels(env);
      t_end = second(); Time = t_end - t_start;
      
      if ((lpstat = CPXgetstat(env, lp))) {
         printf("\n   Exact solution status = %d", lpstat);
         fflush(stdout);
      } else { printf("\nSomething wrong in exact solution: lpstat = %d", lpstat); exit(0); }

      printf("\n   EXACT step performed:"); fflush(stdout);
      switch(lpstat) {
      case 101:
                xBarValue = getSolution(xBar);
                printf("  OPTIMAL sol found %16.4f - elapsed time %8.1f", xBarValue, Time); fflush(stdout);
                break;
      case 102:
                xBarValue = getSolution(xBar);
                printf("  OPTIMAL sol found %16.4f - elapsed time %8.1f", xBarValue, Time); fflush(stdout);
                break;
      case 103:
                printf("  OPTIMAL sol proved %16.4f - elapsed time %8.1f", overallBestValue, Time); fflush(stdout);                break;
      case 107:
                xBarValue = getSolution(xBar);
                printf("  IMPROVED sol %16.4f - elapsed time %8.1f", xBarValue, timeLimit); fflush(stdout);
                break;
      case 108:
                printf("  NO IMPROVED sol %16.4f - elapsed time %8.1f", overallBestValue, timeLimit); fflush(stdout);
                break;
      }
   } else {
      if (!Optimum) printf("\n Final bound = %16.4f after %d B&B nodes", bestBound, nNodes); fflush(stdout);
      printf("\n Final Solution = %16.4f after %8.1f sec and %d(/%d) div", 
                 overallBestValue, time4Best, dBest, dv); fflush(stdout);

      if (Time > timeLimit) printf("\n Time Limit of %8.1f reached", timeLimit);
      else if (nIter == maxIter) printf("\n Max Iteration limit of %d reached", maxIter);
   }

   free(xBar); free(ctype); free(slot);
   printf("\n");

   CPXcloseCPLEX(&env);
   return 0;
}

static int CPXPUBLIC mycallback (CPXENVptr env, void *cbdata, int wherefrom, void *cbhandle)
{
   int    status = 0;
   int    next;
   double dnext, t_end, Time;
   double Value = dInf;

   if (wherefrom == CPX_CALLBACK_MIP) {
      status = CPXgetcallbackinfo(env, cbdata, wherefrom, CPX_CALLBACK_INFO_BEST_INTEGER, &Value);
      if (status) printf("\n Error in CPXgetcallbackinfo: %d", status);

      t_end = second();
      Time = t_end - t_start;

      if (MinMax == 1) {
	 if (Value < overallBestValue - Eps) {
            printf(" NEW INTEGER SOLUTION %16.4f - elapsed time %8.1f\n", Value, Time);
            fflush(stdout);
            overallBestValue = Value;
            time4Best = Time;
            dBest = dv;
            flag = 1;
	 }
      } else {
	 if (Value > overallBestValue + Eps) {
	     printf(" NEW INTEGER SOLUTION %16.4f - elapsed time %8.1f\n", Value, Time);
             fflush(stdout);
             overallBestValue = Value;
             time4Best = Time;
             dBest = dv;
             flag = 1;
	 }  
      }

      dnext = floor(Time/timeSlot);
      next = (int) dnext;
      if (!slot[next]) {
         printf("     BEST KNOWN SOLUTION  %16.4f - elapsed Time %8.1f\n", overallBestValue, Time);
         fflush(stdout);
         slot[next] = 1;
      }
   }
   return (status);
}
