#include "LocBra.h"
#include "LocBra1.h"

void modifyRow(int type)
{
   int        status, lastConstIndex, numrows;
   char       sense;
   double     rhs;

   int       *con2ch, *delete;

   lastConstIndex = CPXgetnumrows(env, lp) - 1;
   status = CPXgetrhs(env, lp, &rhs, lastConstIndex, lastConstIndex);
   con2ch = (int*)calloc(2, sizeof(int));
   con2ch[0] = lastConstIndex;

   switch (type) {
   case 0: 
           sense = 'G';
           rhs += 1.;
           status = CPXchgsense(env, lp, 1, con2ch, &sense);
           if (status) printf("\nError in CPXchgsense: %d", status);
           status = CPXchgrhs(env, lp, 1, con2ch, &rhs);
           if (status) printf("\nError in CPXchgrhs: %d", status);
           break;
   case 1:
           sense = 'G';
           rhs = - lastCardS + 1;
           status = CPXchgsense(env, lp, 1, con2ch, &sense);
           if (status) printf("\nError in CPXchgsense: %d", status);
           status = CPXchgrhs(env, lp, 1, con2ch, &rhs);
           if (status) printf("\nError in CPXchgrhs: %d", status);
           if (original) {
              int j, toAdd;
#ifdef LB
              toAdd = nBcols - lastCardS;
              for (j = 0; j < toAdd; j++) 
                status = CPXchgcoef(env, lp, lastConstIndex, ref[j], 1.);
#else
              toAdd = lastCardS;
              for (j = 0; j < toAdd; j++) 
                status = CPXchgcoef(env, lp, lastConstIndex, ref[j], -1.);
#endif 
           }          
           break;
   case 2:
           numrows = lastConstIndex + 1;
           delete = (int*)calloc(numrows, sizeof(int));
           delete[lastConstIndex] = 1;
           status = CPXdelsetrows(env, lp, delete);
           free(delete);
           break;
   }
   free(con2ch);

} 
 
void localBranch(char *ctype, double *xBar, int refine, double givenRhs) 
{
   int        i, count, status, size, cardS, cnt;
   char       sense;
   double     rhs;

   int       *rmatind, *rmatbeg, *tmp;
   double    *rmatval;
 
   tmp = (int*)calloc(ncols, sizeof(int));
 
   cardS = 0;
   for (i = 0; i < ncols; i++) {
       if (ctype[i] == 'B') {
          if (xBar[i] > 1. - Eps) {
             tmp[i] = -1;
             cardS++;
          } else tmp[i] = 1;
       }
   } 

   size = nBcols;
   if (original && !refine) size = cardS;

   /* used to store the cardinality of the support of the current sol */ 
   lastCardS = cardS; 

   rmatval = (double*)calloc(size, sizeof(double));
   rmatind = (int*)calloc(size, sizeof(int));
   rmatbeg = (int*)calloc(2, sizeof(int));

   count = cnt = 0;
   for (i = 0; i < ncols; i++) {
       if (tmp[i] != 0) {
          if (original && !refine) {
#ifdef LB
             if (tmp[i] < 0) {
#else
             if (tmp[i] > 0) {
#endif
                rmatval[count] = (double) tmp[i];
                rmatind[count] = i;
                count++;
             }
          } else {
             rmatval[count] = (double) tmp[i];
             rmatind[count] = i;
             count++;
             if (original && refine) {
#ifdef LB
                if (tmp[i] == 1)  { ref[cnt] = i; cnt++; }
#else
                if (tmp[i] == -1) { ref[cnt] = i; cnt++; }
#endif
             }
          }
       }
   }

   if (original && !refine) {
#ifdef LB
      if (count != cardS) {
#else
      if (count != nBcols - cardS) {
#endif
         printf("\n error in constraint building!!: count %d -- cardS %d", count, cardS);
         exit(1);
      }
   } else {
      if (count != nBcols) {
         printf("\n error in constraint building!!: count %d -- nBcols %d", count, nBcols);
         exit(1);
      }
   }

   rmatbeg[0] = 0; 

   if (refine) sense = 'E';
   else        sense = 'L'; 
   if (original) {
#ifdef LB
      rhs = givenRhs - cardS;
#else
      rhs = givenRhs;
      if (refine) rhs -= cardS;
#endif
   } else rhs = givenRhs - cardS;

   printf("\n   nBcols = %d, cardS = %d -- rhs = %6.1f", nBcols, cardS, rhs);

   status = CPXaddrows(env, lp, 0, 1, size, &rhs, &sense, rmatbeg, rmatind, rmatval, NULL, NULL);
   if (status) printf("\n error in CPXaddrows: %d", status);

   free(rmatbeg); free(rmatind); free(rmatval); free(tmp);
}

int checkSol(double xBarValue, double *bestValue, double *xBar)
{
   int       Improved = 0;
   int       i;
   double    t_end;

   switch (MinMax) {
   case 1:
            if (xBarValue < *bestValue) {
               *bestValue = xBarValue;
               if (overallBestValue - Eps > *bestValue) {
                  overallBestValue = *bestValue;
                  dBest = dv;
                  t_end = second();
                  time4Best = t_end - t_start; 
		  flag = 1;
               }
	       if (flag) {
                  nI = 0; 
		  for (i = ncols-nNewCols; i < ncols; i++) if (fabs(xBar[i]) > Eps) nI++;
		  if (fabs(overallBestValue-bestBound) < Eps) {
                     printf("\n Optimal solution found");
		     Optimum = 1;
		  }
		  flag = 0;
               }
               Improved = 1;
            }
            break;
   case -1:
            if (xBarValue > *bestValue) {
               *bestValue = xBarValue;
               if (overallBestValue + Eps < *bestValue) {
                  overallBestValue = *bestValue;
                  dBest = dv;
                  t_end = second();
                  time4Best = t_end - t_start; 
		  flag = 1;
               }
	       if (flag) {
                  nI = 0; 
		  for (i = ncols-nNewCols; i < ncols; i++) if (fabs(xBar[i]) > Eps) nI++;
		  if (fabs(overallBestValue-bestBound) < Eps) {
                     printf("\n Optimal solution found");
		     Optimum = 1;
		  }
		  flag = 0;
               }
               Improved = 1;
            }
            break;
   }
   return(Improved);
}

double getSolution(double *xBar)
{
   int       status;
   double    value;

   status = CPXgetmipobjval(env, lp, &value);
   if (status) printf("\nError in CPXgetmipobjval: %d", status);

   status = CPXgetmipx(env, lp, xBar, 0, ncols - 1);
   if (status) printf("\nError in CPXgetx: %d", status);

   return(value);
}

void refineSolution(double *xBarValue, double *xBar)
{
   int       status, refine, lpstat, numrows;
   double    rhs;

   int      *delete;
   char     *ctype;

   ctype  = (char*)calloc(ncols, sizeof(char));
   status = CPXgetctype(env, lp, ctype, 0, ncols - 1);
   if (status) printf("\nError in CPXgetctype: %d", status);

   refine = 1; rhs = 0.;
   localBranch(ctype, xBar, refine, rhs);

   status = CPXsetdblparam(env, CPX_PARAM_TILIM, dInf);
   status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, Inf);
   switch(MinMax) {
   case 1:  /* MINimization version */
	    status = CPXsetdblparam(env, CPX_PARAM_CUTUP, *xBarValue - Eps);
	    break;
   case -1: /* MAXimization version */
	    status = CPXsetdblparam(env, CPX_PARAM_CUTLO, *xBarValue + Eps);
	    break;
   }

   status = CPXmipopt(env, lp);
   if (status) printf("\nError in CPXmipopt: %d", status);

   if ((lpstat = CPXgetstat(env, lp))) printf("\n   Refinement status = %d", lpstat);
   else { 
      printf("\nSomething wrong in refinement: lpstat = %d", lpstat); 
      status = CPXwriteprob(env, lp, "refine.lp", NULL);
      exit(0); 
   }

   switch(lpstat) {
   case 101:
             *xBarValue = getSolution(xBar);
             break;
   case 102:
             *xBarValue = getSolution(xBar);
             break;
   case 103:
             break;
   }

   numrows = CPXgetnumrows(env, lp);
   delete = (int*)calloc(numrows, sizeof(int));
   delete[numrows-1] = 1;
   status = CPXdelsetrows(env, lp, delete);
   free(delete);

   free(ctype); 
}
